# SPDX-FileCopyrightText: 2024 Julien Rippinger
#
# SPDX-License-Identifier: GPL-3.0-or-later


import random

from compas.geometry import Line as CLine
from compas.geometry import Point as CPoint
from compas.geometry import intersection_line_line_xy

from axonometry.utils import pair_projections_lines

from .config import config
from .point import Point, is_coplanar, random_point


class Line:
    """The basic primitive for a drawing.

    :param start: The first point.
    :param end: The second point.
    """

    def __init__(self, start: Point, end: Point) -> None:  # noqa: D107
        assert is_coplanar([start, end]), "Points are not in the same plane."
        self.plane = None  # set by parent
        self.projections = {
            "xy": None,
            "yz": None,
            "zx": None,
            "xyz": [],
        }  #: Collection of projections
        self.start = start
        self.end = end
        self.key = {start.key, end.key}.pop()
        self.data: CLine | None = None

    @property
    def __data__(self):
        return {"start": self.start.data.__data__, "end": self.end.data.__data__}

    def __iter__(self):
        """Necessary for compas.geometry functions"""
        return iter([self.start.data, self.end.data])

    def project(
        self,
        distance: float | None = None,
        ref_plane_key: str | None = None,
    ) -> "Line":
        """Project crurrent line on another plane."""
        # determine projection origin plane
        assert self.plane, "To project, the line needs to be part of a plane."
        if self.plane.key == "xyz":
            new_line = self.__project_on_reference_plane(ref_plane_key)
        else:
            new_line = self.__project_on_axonometry_plane(distance)

        return new_line

    def __project_on_axonometry_plane(
        self,
        distance: float,
        start_distance: float | None = None,
        end_distance: float | None = None,
    ) -> "Line":
        """Project line on the axonometric picture plane.

        :param distance: The third coordinate value relative to the current reference plane.
        :param start_distance: Fine-tuning specific distance for beginning of line.
        :param end_distance: Fine-tuning specific distance for end of line.
        """
        start_distance = start_distance if start_distance else distance
        end_distance = end_distance if end_distance else distance

        if self.plane.key == "xy":
            new_line = Line(
                Point(x=self.start.x, y=self.start.y, z=start_distance),
                Point(x=self.end.x, y=self.end.y, z=end_distance),
            )  # data will be update
            ref_plane_key = random.choice(["yz", "zx"])  # noqa: S311
            if ref_plane_key == "yz":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(y=self.start.y, z=start_distance),
                        Point(y=self.end.y, z=end_distance),
                    ),
                )
            elif ref_plane_key == "zx":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(x=self.start.x, z=start_distance),
                        Point(x=self.end.x, z=end_distance),
                    ),
                )

        elif self.plane.key == "yz":
            new_line = Line(
                Point(x=start_distance, y=self.start.y, z=self.start.z),
                Point(x=end_distance, y=self.end.y, z=self.end.z),
            )  # data will be update
            ref_plane_key = "xy"  # random.choice(["zx", "xy"])
            if ref_plane_key == "zx":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(z=self.start.z, x=start_distance),
                        Point(z=self.end.z, x=end_distance),
                    ),
                )
            elif ref_plane_key == "xy":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(y=self.start.y, x=start_distance),
                        Point(y=self.end.y, x=end_distance),
                    ),
                )

        elif self.plane.key == "zx":
            new_line = Line(
                Point(x=self.start.x, y=start_distance, z=self.start.z),
                Point(x=self.end.x, y=end_distance, z=self.end.z),
            )  # data will be update
            ref_plane_key = "xy"  # random.choice(["xy", "yz"])
            if ref_plane_key == "xy":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(x=self.start.x, y=start_distance),
                        Point(x=self.end.x, y=end_distance),
                    ),
                )
            elif ref_plane_key == "yz":
                auxilary_line = self.plane.axo[ref_plane_key].add_line(
                    Line(
                        Point(x=self.start.x, y=start_distance),
                        Point(z=self.end.z, y=end_distance),
                    ),
                )

        axo_line_start_data = intersection_line_line_xy(
            CLine.from_point_and_vector(self.start.data, self.plane.projection_vector),
            CLine.from_point_and_vector(
                auxilary_line.start.data,
                self.plane.axo[ref_plane_key].projection_vector,
            ),
        )

        axo_line_end_data = intersection_line_line_xy(
            CLine.from_point_and_vector(self.end.data, self.plane.projection_vector),
            CLine.from_point_and_vector(
                auxilary_line.end.data,
                self.plane.axo[ref_plane_key].projection_vector,
            ),
        )

        new_line.data = CLine(CPoint(*axo_line_start_data), CPoint(*axo_line_end_data))
        new_line.start.data = new_line.data[0]
        new_line.end.data = new_line.data[1]

        self.plane.drawing.add_compas_geometry(
            [
                CLine(self.start.data, axo_line_start_data),
                CLine(auxilary_line.start.data, axo_line_start_data),
                CLine(self.end.data, axo_line_end_data),
                CLine(auxilary_line.end.data, axo_line_end_data),
            ],
        )

        self.plane.axo.add_line(new_line)
        pair_projections_lines(new_line, auxilary_line)
        pair_projections_lines(new_line, self)

        return new_line

    def __project_on_reference_plane(self, ref_plane_key: str) -> "Line":
        """Project the line on the selected reference plane.

        Knowing that the current line is member of the axo plane.
        """
        if self == self.projections[ref_plane_key]:
            # projection of line already exists, nothing to do
            config.logger.info("Line already exists.")
            new_line = self.projections[ref_plane_key]
        else:
            if ref_plane_key == "xy":
                new_line = self.plane[ref_plane_key].add_line(
                    Line(
                        Point(x=self.start.x, y=self.start.y),
                        Point(x=self.end.x, y=self.end.y),
                    ),
                )
            elif ref_plane_key == "yz":
                new_line = self.plane[ref_plane_key].add_line(
                    Line(
                        Point(y=self.start.y, z=self.start.z),
                        Point(y=self.end.y, z=self.end.z),
                    ),
                )
            elif ref_plane_key == "zx":
                new_line = self.plane[ref_plane_key].add_line(
                    Line(
                        Point(x=self.start.x, z=self.start.z),
                        Point(x=self.end.x, z=self.end.z),
                    ),
                )

            # draw new projection line
            self.plane.drawing.add_compas_geometry(
                [
                    CLine(self.start.data, new_line.start.data),
                    CLine(self.end.data, new_line.end.data),
                ],
            )

            pair_projections_lines(self, new_line)

        return new_line

    def __eq__(self, other: "Line") -> bool:
        """Lines are equal when their points are equal."""
        if (not isinstance(other, type(self))) or (self is None or other is None):
            # if the other item of comparison is not also of the Point class
            return False
        return (self.start == other.start and self.end == other.end) or (
            self.start == other.end and self.end == other.start
        )

    def __repr__(self):
        """Made of points."""
        return f"Line({self.start}, {self.end})"


def random_line(key: str = "xyz") -> Line:
    if key == "xy":
        return Line(random_point("xy"), random_point("xy"))
    if key == "yz":
        return Line(random_point("yz"), random_point("yz"))
    if key == "zx":
        return Line(random_point("zx"), random_point("zx"))
    return Line(random_point(), random_point())
