# SPDX-FileCopyrightText: 2024 Julien Rippinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import TYPE_CHECKING

from compas.geometry import Line as CLine
from compas.geometry import Point as CPoint
from compas.geometry import Vector as CVector
from compas.geometry import intersection_line_line_xy

from .config import config
from .line import Line
from .point import Point
from .utils import pair_projections_lines, pair_projections_points, random_axo_ref_plane_keys

if TYPE_CHECKING:
    from .trihedron import Trihedron


class Plane:
    """Base class for axonometric and reference planes."""

    def __init__(self) -> None:  # noqa: D107
        self.key = None
        self.drawing = None  # Initialize the drawing attribute if needed
        self.objects = []  # TODO: make dict with type as key

    def add_point(self, point: Point, ref_plane_keys: list[str] | None = None) -> Point:
        """Add a point to the current plane."""
        assert point.key == self.key, (
            f"Point coordinates must follow containing plane coordinates. Plane:{self.key} & Point:{point.key}"
        )
        if self.key == "xyz":
            # Point data could not exist
            config.logger.debug(
                f"[{self.key.upper()}] Adding {point} by auxilary projection.",
            )
            if point.data is None:
                point.data = self.__decompose_axo_point(point, ref_plane_keys)

        else:
            if point.matrix_applied:
                # Reuse the original when repeating operation.
                point.reset_data()
            point.data = point.data.transformed(self.matrix)
            point.matrix_applied = True

        config.logger.debug(f"[{self.key.upper()}] Add {point}")
        self.objects.append(point)
        self.drawing.add(point)
        point.plane = self  # add self as parent
        config.logger.debug(f"[{self.key.upper()}] Current objects in {self}: {self.objects}.")
        return point

    def add_line(self, line: Line, ref_plane_keys: list[str] | None = None) -> Line:
        """Add a line to the current reference plane.

        These are the various cases:
        In XYZ space:
            - The line has no data: it's a new line and auxilary projections are necessary.
            - The line has data set: data was computed by the projection function.
        In a reference plane:
            -
        """
        assert line.key == self.key, (
            f"Line coordinates must coorespond to plane coordinates system. Plane:{self.key=} & Point:{line.key=}"
        )

        line.plane = self

        if self.key == "xyz":
            if line.data is None:
                config.logger.debug(
                    f"[{self.key.upper()}] Adding {line} by auxilary projection.",
                )
                if not ref_plane_keys:
                    ref_plane_keys = random_axo_ref_plane_keys()
                # Add points to axonometry space.
                # This will trigger the decomposition in the add_point funciton.
                line.start = self.add_point(line.start, ref_plane_keys=ref_plane_keys)
                line.end = self.add_point(line.end, ref_plane_keys=ref_plane_keys)
                # Update the line data with the projection
                line.data = CLine(line.start.data, line.end.data)

                # Use auxilary lines if data was not yet computed
                for ref_plane_key in self.__common_projections(
                    line.start.projections,
                    line.end.projections,
                ):
                    auxilary_line = Line(
                        line.start.projections[ref_plane_key],
                        line.end.projections[ref_plane_key],
                    )
                    self[ref_plane_key].add_line(auxilary_line)
                    pair_projections_lines(line, auxilary_line)

        else:
            if not ref_plane_keys:
                ref_plane_keys = random_axo_ref_plane_keys()
            # Compute the projected points of the line
            line.start = self.add_point(line.start, ref_plane_keys=ref_plane_keys)
            line.end = self.add_point(line.end, ref_plane_keys=ref_plane_keys)
            # Update the line data with the projection
            line.data = CLine(line.start.data, line.end.data)

        config.logger.info(f"[{self.key.upper()}] Add {line}")
        self.objects.append(line)
        self.drawing.add(line)

        return line

    def __common_projections(self, dict1, dict2):
        """Find which projected points are on the same reference plane."""
        for key in dict1:
            if key == "xyz":  # Exclude this specific key from comparison
                continue
            if key in dict2 and dict1[key] is not None and dict2[key] is not None:
                yield key

    def __decompose_axo_point(
        self,
        axo_point: Point,
        ref_plane_keys: list[str] | None = None,
    ) -> CPoint:
        """Directly added point in XYZ space becomes the intersection of two projected points.

        Basically adding points in two reference planes and intersecting them
        in the xyz space. The two planes can be provided as a parameter.

        :param: The axonometric point to be found by intersection of two projections.
        :param ref_plane_keys: The two reference, default to XY and random YZ or ZX.
        :return: Intersection coordinates from projected points.
        """
        config.logger.debug(f"Decompose {axo_point=}")

        if (
            ref_plane_keys
            and len(ref_plane_keys) != 2
            and (
                ref_plane_keys != ["xy", "yz"]
                or ref_plane_keys != ["xy", "zx"]
                or ref_plane_keys != ["yz", "zx"]
            )
        ):
            raise ValueError(f"{ref_plane_keys} are invalid.")

        if not ref_plane_keys:
            ref_plane_keys = random_axo_ref_plane_keys()

        if "xy" in ref_plane_keys and "yz" in ref_plane_keys:
            p1 = Point(x=axo_point.x, y=axo_point.y)
            p2 = Point(y=axo_point.y, z=axo_point.z)
            plane1 = self["xy"]
            plane2 = self["yz"]

        if "zx" in ref_plane_keys and "yz" in ref_plane_keys:
            p1 = Point(y=axo_point.y, z=axo_point.z)
            p2 = Point(z=axo_point.z, x=axo_point.x)
            plane1 = self["yz"]
            plane2 = self["zx"]

        if "xy" in ref_plane_keys and "zx" in ref_plane_keys:
            p1 = Point(z=axo_point.z, x=axo_point.x)
            p2 = Point(x=axo_point.x, y=axo_point.y)
            plane1 = self["zx"]
            plane2 = self["xy"]

        config.logger.debug(f"Two auxilary points computed {p1=}, {p2=}")

        plane1.add_point(p1)
        plane2.add_point(p2)
        pair_projections_points(axo_point, p1)
        pair_projections_points(axo_point, p2)

        # add them in respective ReferencePlanes
        axo_point_data = intersection_line_line_xy(
            CLine.from_point_and_vector(p1.data, plane1.projection_vector),
            CLine.from_point_and_vector(p2.data, plane2.projection_vector),
        )
        axo_point_data = CPoint(*axo_point_data)
        config.logger.debug(f"New {axo_point_data=}")
        # Add points in reference planes to the
        # axo point projections collection

        # draw intersection
        self.drawing.add_compas_geometry(
            [CLine(p1.data, axo_point_data), CLine(p2.data, axo_point_data)],
        )
        return axo_point_data


class ReferencePlane(Plane):
    """Represents a reference plane in an axonometric projection.

    :param lines: The two lines making up the reference plane axes.

    """

    def __init__(self, line_pair: list[CLine], projection_vector: CVector) -> None:  # noqa: D107
        super().__init__()  # Call the parent class constructor if necessary
        self.trihedron: Trihedron | None = None
        self.matrix = None
        self.axes = line_pair
        self.projection_vector = projection_vector
        self.matrix_to_coord_plane = None  # TODO

    def __repr__(self) -> str:
        """Get axes keys."""
        return f"Reference Plane {self.key.upper()}"

    def add_svg_file(self, svg_file: str):
        """Get an external svg and add it to current reference plane.

        An SVG is treated as a collection of lines. The steps to follow are extracting the line
        coordinates and adding each line to the current plane. Roughly the code should be as
        follow::

            for line in collection:
                self.add_line(Line(line))  # this will call the matrix
            doc = self.drawing.convert_svg_vpype_doc(svg_file)
        """
        raise NotImplementedError
